import importlib
from unittest import IsolatedAsyncioTestCase

from marshmallow import ValidationError

from .....admin.request_context import AdminRequestContext
from .....anoncreds.models.presentation_request import (
    AnoncredsPresentationReqAttrSpecSchema,
)
from .....indy.holder import IndyHolder
from .....indy.verifier import IndyVerifier
from .....ledger.base import BaseLedger
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import routes as test_module


class TestProofRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context({}, profile=self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_validate_proof_req_attr_spec(self):
        aspec = AnoncredsPresentationReqAttrSpecSchema()
        aspec.validate_fields({"name": "attr0"})
        aspec.validate_fields(
            {
                "names": ["attr0", "attr1"],
                "restrictions": [{"attr::attr1::value": "my-value"}],
            }
        )
        aspec.validate_fields(
            {"name": "attr0", "restrictions": [{"schema_name": "preferences"}]}
        )
        with self.assertRaises(ValidationError):
            aspec.validate_fields({})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"name": "attr0", "names": ["attr1", "attr2"]})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr1", "attr2"]})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr0", "attr1"], "restrictions": []})
        with self.assertRaises(ValidationError):
            aspec.validate_fields({"names": ["attr0", "attr1"], "restrictions": [{}]})

    async def test_presentation_exchange_list(self):
        self.request.query = {
            "thread_id": "thread_id_0",
            "connection_id": "conn_id_0",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.query = mock.CoroutineMock()
            mock_presentation_exchange.query.return_value = [mock_presentation_exchange]
            mock_presentation_exchange.serialize = mock.MagicMock()
            mock_presentation_exchange.serialize.return_value = {
                "thread_id": "sample-thread-id"
            }

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_list(self.request)
                mock_response.assert_called_once_with(
                    {"results": [mock_presentation_exchange.serialize.return_value]}
                )

    async def test_presentation_exchange_list_x(self):
        self.request.query = {
            "thread_id": "thread_id_0",
            "connection_id": "conn_id_0",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.query = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_list(self.request)

    async def test_presentation_exchange_credentials_list_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock()

            # Emulate storage not found (bad presentation exchange id)
            mock_presentation_exchange.retrieve_by_id.side_effect = StorageNotFoundError

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.presentation_exchange_credentials_list(self.request)

    async def test_presentation_exchange_credentials_x(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1",
        }
        self.request.query = {"extra_query": {}}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(side_effect=test_module.IndyHolderError())
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)
        mock_px_rec = mock.MagicMock(save_error_state=mock.CoroutineMock())

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id.return_value = mock_px_rec

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_credentials_list(self.request)

    async def test_presentation_exchange_credentials_list_single_referent(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [{"name": "Credential1"}, {"name": "Credential2"}]
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id.return_value = mock.MagicMock()

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_credentials_list(self.request)
                mock_response.assert_called_once_with(returned_credentials)

    async def test_presentation_exchange_credentials_list_multiple_referents(self):
        self.request.match_info = {
            "pres_ex_id": "123-456-789",
            "referent": "myReferent1,myReferent2",
        }
        self.request.query = {"extra_query": {}}

        returned_credentials = [{"name": "Credential1"}, {"name": "Credential2"}]
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials_for_presentation_request_by_referent = (
            mock.CoroutineMock(return_value=returned_credentials)
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_credentials_list(self.request)
                mock_response.assert_called_once_with(returned_credentials)

    async def test_presentation_exchange_retrieve(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_pres_ex:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_pres_ex.retrieve_by_id = mock.CoroutineMock()
            mock_pres_ex.retrieve_by_id.return_value = mock_pres_ex
            mock_pres_ex.serialize = mock.MagicMock()
            mock_pres_ex.serialize.return_value = {"thread_id": "sample-thread-id"}

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_retrieve(self.request)
                mock_response.assert_called_once_with(mock_pres_ex.serialize.return_value)

    async def test_presentation_exchange_retrieve_not_found(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_pres_ex:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_pres_ex.retrieve_by_id = mock.CoroutineMock()

            # Emulate storage not found (bad presentation exchange id)
            mock_pres_ex.retrieve_by_id.side_effect = StorageNotFoundError

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.presentation_exchange_retrieve(self.request)

    async def test_presentation_exchange_retrieve_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        mock_pres_ex_rec = mock.MagicMock(
            connection_id="abc123",
            thread_id="thid123",
            save_error_state=mock.CoroutineMock(),
        )
        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_pres_ex:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_pres_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_pres_ex_rec
            )
            mock_pres_ex_rec.serialize = mock.MagicMock(
                side_effect=test_module.BaseModelError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_retrieve(self.request)

    async def test_presentation_exchange_send_proposal(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ) as mock_preview,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange_record = mock.MagicMock()
            mock_presentation_manager.return_value.create_exchange_for_proposal = (
                mock.CoroutineMock(return_value=mock_presentation_exchange_record)
            )

            mock_preview.return_value.deserialize.return_value = mock.MagicMock()

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_send_proposal(self.request)
                mock_response.assert_called_once_with(
                    mock_presentation_exchange_record.serialize.return_value
                )

    async def test_presentation_exchange_send_proposal_no_conn_record(self):
        self.request.json = mock.CoroutineMock()

        with mock.patch(
            "acapy_agent.connections.models.conn_record.ConnRecord",
            autospec=True,
        ) as mock_connection_record:
            # Since we are mocking import
            importlib.reload(test_module)

            # Emulate storage not found (bad connection id)
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_proposal(self.request)

    async def test_presentation_exchange_send_proposal_not_ready(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "messages.presentation_proposal.PresentationProposal"
                ),
                autospec=True,
            ),
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_connection_record.retrieve_by_id = mock.CoroutineMock()
            mock_connection_record.retrieve_by_id.return_value.is_ready = False

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.presentation_exchange_send_proposal(self.request)

    async def test_presentation_exchange_send_proposal_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_manager.return_value.create_exchange_for_proposal = (
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                        save_error_state=mock.CoroutineMock(),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_proposal(self.request)

    async def test_presentation_exchange_create_request(self):
        self.request.json = mock.CoroutineMock(
            return_value={"comment": "dummy", "proof_request": {}}
        )

        with (
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ) as mock_attach_decorator,
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )
            mock_presentation_exchange.serialize = mock.MagicMock()
            mock_presentation_exchange.serialize.return_value = {
                "thread_id": "sample-thread-id"
            }
            mock_mgr = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock_presentation_exchange
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_create_request(self.request)
                mock_response.assert_called_once_with(
                    mock_presentation_exchange.serialize.return_value
                )

    async def test_presentation_exchange_create_request_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={"comment": "dummy", "proof_request": {}}
        )

        with (
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_manager.return_value.create_exchange_for_request = (
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                        save_error_state=mock.CoroutineMock(),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_create_request(self.request)

    async def test_presentation_exchange_send_free_request(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": "dummy",
                "comment": "dummy",
                "proof_request": {},
            }
        )

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ) as mock_attach_decorator,
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )
            mock_presentation_exchange.serialize = mock.MagicMock()
            mock_presentation_exchange.serialize.return_value = {
                "thread_id": "sample-thread-id"
            }

            mock_mgr = mock.MagicMock(
                create_exchange_for_request=mock.CoroutineMock(
                    return_value=mock_presentation_exchange
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_send_free_request(self.request)
                mock_response.assert_called_once_with(
                    mock_presentation_exchange.serialize.return_value
                )

    async def test_presentation_exchange_send_free_request_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"connection_id": "dummy"})

        with mock.patch(
            "acapy_agent.connections.models.conn_record.ConnRecord",
            autospec=True,
        ) as mock_connection_record:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_connection_record.retrieve_by_id = mock.CoroutineMock()
            mock_connection_record.retrieve_by_id.side_effect = StorageNotFoundError

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_free_request(self.request)

    async def test_presentation_exchange_send_free_request_not_ready(self):
        self.request.json = mock.CoroutineMock(
            return_value={"connection_id": "dummy", "proof_request": {}}
        )

        with mock.patch(
            "acapy_agent.connections.models.conn_record.ConnRecord",
            autospec=True,
        ) as mock_connection_record:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_connection_record.is_ready = False
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.presentation_exchange_send_free_request(self.request)

    async def test_presentation_exchange_send_free_request_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": "dummy",
                "comment": "dummy",
                "proof_request": {},
            }
        )

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ) as mock_attach_decorator,
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ),
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_manager.return_value.create_exchange_for_request = (
                mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(side_effect=test_module.StorageError()),
                        save_error_state=mock.CoroutineMock(),
                    )
                )
            )

            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_attach_decorator.data_base64 = mock.MagicMock(
                return_value=mock_attach_decorator
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_free_request(self.request)

    async def test_presentation_exchange_send_bound_request(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(
                test_module, "PresentationRequest", autospec=True
            ) as mock_presentation_request,
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange",
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.state = (
                test_module.V10PresentationExchange.STATE_PROPOSAL_RECEIVED
            )
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_presentation_exchange
            )
            mock_presentation_exchange.serialize = mock.MagicMock()
            mock_presentation_exchange.serialize.return_value = {
                "thread_id": "sample-thread-id"
            }
            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )

            mock_mgr = mock.MagicMock(
                create_bound_request=mock.CoroutineMock(
                    return_value=(mock_presentation_exchange, mock_presentation_request)
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_send_bound_request(self.request)
                mock_response.assert_called_once_with(
                    mock_presentation_exchange.serialize.return_value
                )

    async def test_presentation_exchange_send_bound_request_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.state = (
                test_module.V10PresentationExchange.STATE_PROPOSAL_RECEIVED
            )
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_presentation_exchange
            )

            mock_connection_record.retrieve_by_id = mock.CoroutineMock()
            mock_connection_record.retrieve_by_id.side_effect = StorageNotFoundError

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_bound_request(self.request)

    async def test_presentation_exchange_send_bound_request_not_ready(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.state = (
                test_module.V10PresentationExchange.STATE_PROPOSAL_RECEIVED
            )
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_presentation_exchange
            )

            mock_connection_record.is_ready = False
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.presentation_exchange_send_bound_request(self.request)

    async def test_presentation_exchange_send_bound_request_px_rec_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module.V10PresentationExchange,
            "retrieve_by_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = StorageNotFoundError("no such record")
            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.presentation_exchange_send_bound_request(self.request)
            assert "no such record" in str(context.exception)

    async def test_presentation_exchange_send_bound_request_bad_state(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_PRESENTATION_ACKED
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_bound_request(self.request)

    async def test_presentation_exchange_send_bound_request_x(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.state = (
                test_module.V10PresentationExchange.STATE_PROPOSAL_RECEIVED
            )
            mock_presentation_exchange.connection_id = "abc123"
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_presentation_exchange
            )
            mock_presentation_exchange.serialize = mock.MagicMock()
            mock_presentation_exchange.serialize.return_value = {
                "thread_id": "sample-thread-id",
            }
            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )

            mock_mgr = mock.MagicMock(
                create_bound_request=mock.CoroutineMock(
                    side_effect=[
                        test_module.LedgerError(),
                        test_module.StorageError(),
                    ]
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with self.assertRaises(test_module.web.HTTPBadRequest):  # ledger error
                await test_module.presentation_exchange_send_bound_request(self.request)
            with self.assertRaises(test_module.web.HTTPBadRequest):  # storage error
                await test_module.presentation_exchange_send_bound_request(self.request)

    async def test_presentation_exchange_send_presentation(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "dummy",
                "self_attested_attributes": {},
                "requested_attributes": {},
                "requested_predicates": {},
            }
        )
        self.request.match_info = {"pres_ex_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.state = (
                test_module.V10PresentationExchange.STATE_REQUEST_RECEIVED
            )
            mock_presentation_exchange.connection_id = "dummy"
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_REQUEST_RECEIVED,
                    connection_id="dummy",
                    serialize=mock.MagicMock(
                        return_value={"thread_id": "sample-thread-id"}
                    ),
                )
            )
            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_mgr = mock.MagicMock(
                create_presentation=mock.CoroutineMock(
                    return_value=(mock_presentation_exchange, mock.MagicMock())
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_send_presentation(self.request)
                mock_response.assert_called_once_with(
                    mock_presentation_exchange.serialize.return_value
                )

    async def test_presentation_exchange_send_presentation_px_rec_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module.V10PresentationExchange,
            "retrieve_by_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = StorageNotFoundError("no such record")
            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.presentation_exchange_send_presentation(self.request)
            assert "no such record" in str(context.exception)

    async def test_presentation_exchange_send_presentation_not_found(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_REQUEST_RECEIVED,
                    connection_id="dummy",
                )
            )

            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_presentation(self.request)

    async def test_presentation_exchange_send_presentation_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_REQUEST_RECEIVED,
                    connection_id="dummy",
                )
            )

            mock_connection_record.is_ready = False
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.presentation_exchange_send_presentation(self.request)

    async def test_presentation_exchange_send_presentation_bad_state(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_PRESENTATION_ACKED
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_presentation(self.request)

    async def test_presentation_exchange_send_presentation_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "dummy",
                "self_attested_attributes": {},
                "requested_attributes": {},
                "requested_predicates": {},
            }
        )
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch.object(test_module, "IndyPresPreview", autospec=True),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_REQUEST_RECEIVED,
                    connection_id="dummy",
                    serialize=mock.MagicMock(
                        return_value={"thread_id": "sample-thread-id"}
                    ),
                    save_error_state=mock.CoroutineMock(),
                ),
            )
            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_mgr = mock.MagicMock(
                create_presentation=mock.CoroutineMock(
                    side_effect=test_module.LedgerError()
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_send_presentation(self.request)

    async def test_presentation_exchange_verify_presentation(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                "acapy_agent.indy.util.generate_pr_nonce",
                autospec=True,
            ),
            mock.patch(
                "acapy_agent.indy.models.pres_preview.IndyPresPreview",
                autospec=True,
            ),
            mock.patch.object(test_module, "PresentationRequest", autospec=True),
            mock.patch(
                "acapy_agent.messaging.decorators.attach_decorator.AttachDecorator",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_PRESENTATION_RECEIVED,
                    connection_id="dummy",
                    thread_id="dummy",
                    serialize=mock.MagicMock(
                        return_value={"thread_id": "sample-thread-id"}
                    ),
                )
            )
            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_mgr = mock.MagicMock(
                verify_presentation=mock.CoroutineMock(
                    return_value=mock_presentation_exchange.retrieve_by_id.return_value
                )
            )
            mock_presentation_manager.return_value = mock_mgr

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_verify_presentation(self.request)
                mock_response.assert_called_once_with({"thread_id": "sample-thread-id"})

    async def test_presentation_exchange_verify_presentation_px_rec_not_found(self):
        self.request.json = mock.CoroutineMock(return_value={"trace": False})
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch.object(
            test_module.V10PresentationExchange,
            "retrieve_by_id",
            mock.CoroutineMock(),
        ) as mock_retrieve:
            mock_retrieve.side_effect = StorageNotFoundError("no such record")
            with self.assertRaises(test_module.web.HTTPNotFound) as context:
                await test_module.presentation_exchange_verify_presentation(self.request)
            assert "no such record" in str(context.exception)

    async def test_presentation_exchange_verify_presentation_bad_state(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_PRESENTATION_ACKED
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_verify_presentation(self.request)

    async def test_presentation_exchange_verify_presentation_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseLedger,
            mock.MagicMock(
                __aenter__=mock.CoroutineMock(),
                __aexit__=mock.CoroutineMock(),
            ),
        )
        self.profile.context.injector.bind_instance(
            IndyVerifier,
            mock.MagicMock(
                verify_presentation=mock.CoroutineMock(),
            ),
        )

        with (
            mock.patch(
                "acapy_agent.connections.models.conn_record.ConnRecord",
                autospec=True,
            ) as mock_connection_record,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ) as mock_presentation_manager,
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_presentation_exchange,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_PRESENTATION_RECEIVED,
                    connection_id="dummy",
                    thread_id="dummy",
                    serialize=mock.MagicMock(
                        return_value={"thread_id": "sample-thread-id"}
                    ),
                    save_error_state=mock.CoroutineMock(),
                )
            )

            mock_connection_record.is_ready = True
            mock_connection_record.retrieve_by_id = mock.CoroutineMock(
                return_value=mock_connection_record
            )
            mock_mgr = mock.MagicMock(
                verify_presentation=mock.CoroutineMock(
                    side_effect=[
                        test_module.LedgerError(),
                        test_module.StorageError(),
                    ]
                ),
            )
            mock_presentation_manager.return_value = mock_mgr

            with self.assertRaises(test_module.web.HTTPBadRequest):  # ledger error
                await test_module.presentation_exchange_verify_presentation(self.request)
            with self.assertRaises(test_module.web.HTTPBadRequest):  # storage error
                await test_module.presentation_exchange_verify_presentation(self.request)

    async def test_presentation_exchange_problem_report(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}
        magic_report = mock.MagicMock()

        with (
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_pres_ex,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch.object(
                test_module, "problem_report_for_record", mock.MagicMock()
            ) as mock_problem_report,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_pres_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_problem_report.return_value = magic_report

            await test_module.presentation_exchange_problem_report(self.request)

            self.request["outbound_message_router"].assert_awaited_once()
            mock_response.assert_called_once_with({})

    async def test_presentation_exchange_problem_report_bad_pres_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_pres_ex,
        ):
            # Since we are mocking import
            importlib.reload(test_module)

            mock_pres_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.presentation_exchange_problem_report(self.request)

    async def test_presentation_exchange_problem_report_x(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with (
            mock.patch(
                (
                    "acapy_agent.protocols.present_proof.v1_0."
                    "models.presentation_exchange.V10PresentationExchange"
                ),
                autospec=True,
            ) as mock_pres_ex,
            mock.patch(
                "acapy_agent.protocols.present_proof.v1_0.manager.PresentationManager",
                autospec=True,
            ),
            mock.patch.object(test_module, "problem_report_for_record", mock.MagicMock()),
            mock.patch.object(test_module.web, "json_response"),
        ):
            # Since we are mocking import
            importlib.reload(test_module)
            mock_pres_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_problem_report(self.request)

    async def test_presentation_exchange_remove(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_VERIFIED,
                    connection_id="dummy",
                    delete_record=mock.CoroutineMock(),
                )
            )

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.presentation_exchange_remove(self.request)
                mock_response.assert_called_once_with({})

    async def test_presentation_exchange_remove_not_found(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            # Emulate storage not found (bad pres ex id)
            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                side_effect=StorageNotFoundError
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.presentation_exchange_remove(self.request)

    async def test_presentation_exchange_remove_x(self):
        self.request.match_info = {"pres_ex_id": "dummy"}

        with mock.patch(
            (
                "acapy_agent.protocols.present_proof.v1_0."
                "models.presentation_exchange.V10PresentationExchange"
            ),
            autospec=True,
        ) as mock_presentation_exchange:
            # Since we are mocking import
            importlib.reload(test_module)

            mock_presentation_exchange.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_presentation_exchange.STATE_VERIFIED,
                    connection_id="dummy",
                    delete_record=mock.CoroutineMock(
                        side_effect=test_module.StorageError()
                    ),
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.presentation_exchange_remove(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Module to contain logic to generate the banner for ACA-py."""

import sys
import textwrap
from contextlib import contextmanager
from enum import Enum, auto
from typing import Optional, TextIO


@contextmanager
def Banner(border: str, length: int, file: Optional[TextIO] = None):
    """Context manager to generate a banner for ACA-py."""
    banner = _Banner(border, length, file)
    banner.print_border()
    yield banner
    banner.print_border()


class _Banner:
    """Management class to generate a banner for ACA-py."""

    class align(Enum):
        """Alignment options for banner elements."""

        LEFT = auto()
        CENTER = auto()
        RIGHT = auto()

    def __init__(self, border: str, length: int, file: Optional[TextIO] = None):
        """Initialize the banner object.

        The ``border`` is a single character to be used, and ``length``
        is the desired length of the whole banner, inclusive.
        """
        self.border = border
        self.length = length
        self.file = file or sys.stdout

    def _print(self, text: str):
        """Print value."""
        print(text, file=self.file)

    def _lr_pad(self, content: str):
        """Pad string content with defined border character.

        Args:
            content: String content to pad
        """
        return f"{self.border}{self.border} {content} {self.border}{self.border}"

    def _print_line(self, text: str, alignment: align = align.LEFT):
        """Print line."""
        lines = textwrap.wrap(text, width=self.length)
        for line in lines:
            if len(line) < self.length:
                if alignment == self.align.LEFT:
                    left = ""
                    right = " " * (self.length - len(line))
                elif alignment == self.align.CENTER:
                    left = " " * ((self.length - len(line)) // 2)
                    right = " " * ((self.length - len(line)) // 2)
                    if len(line) % 2 != 0:
                        right += " "
                elif alignment == self.align.RIGHT:
                    left = " " * (self.length - len(line))
                    right = ""
                else:
                    raise ValueError(f"Invalid alignment: {alignment}")
                line = f"{left}{line}{right}"
            self._print(self._lr_pad(line))

    def print_border(self):
        """Print a full line using the border character."""
        self._print(self.border * (self.length + 6))

    def title(self, title, spacing_after: int = 2):
        """Print the main title element."""
        self._print_line(title, self.align.CENTER)
        for _ in range(spacing_after):
            self.spacer()

    def spacer(self):
        """Print an empty line with the border character only."""
        self._print(self._lr_pad(" " * self.length))

    def hr(self, char: str = "-"):
        """Print a line with a horizontal rule."""
        self._print(self._lr_pad(char * self.length))

    def subtitle(self, title: str, spacing_after: int = 1):
        """Print a subtitle for a section."""
        title += ":"
        self._print_line(title, self.align.LEFT)
        for _ in range(spacing_after):
            self.spacer()

    def list(self, items, spacing_after: int = 1):
        """Print a list of items, prepending a dash to each item."""
        for item in items:
            self._print_line(f"  - {item}", self.align.LEFT)

        for _ in range(spacing_after):
            self.spacer()

    def version(self, version):
        """Print the current ``version``."""
        version = f"ver: {version}"
        self._print_line(version, self.align.RIGHT)

    def print(self, text: str):
        """Print a line of text."""
        self._print_line(text, self.align.LEFT)

    def left(self, text: str):
        """Print a line of text left aligned.

        Same as `print` method.
        """
        self._print_line(text, self.align.LEFT)

    def centered(self, text: str):
        """Print a line of text centered."""
        self._print_line(text, self.align.CENTER)

    def right(self, text: str):
        """Print a line of text right aligned."""
        self._print_line(text, self.align.RIGHT)



"""Multiledger related utility methods."""

from collections import OrderedDict

from ..config.settings import BaseSettings
from ..core.error import ProfileError


def get_write_ledger_config_for_profile(settings: BaseSettings) -> dict:
    """Return initial/default write ledger config on profile creation."""
    write_ledger_config = None
    prod_write_ledger_pool = OrderedDict()
    non_prod_write_ledger_pool = OrderedDict()
    for ledger_config in settings.get("ledger.ledger_config_list"):
        if ledger_config.get("is_production") and (
            ledger_config.get("is_write") or settings.get("ledger.read_only")
        ):
            prod_write_ledger_pool[
                ledger_config.get("id") or ledger_config.get("pool_name")
            ] = ledger_config
        elif not ledger_config.get("is_production") and (
            ledger_config.get("is_write") or settings.get("ledger.read_only")
        ):
            non_prod_write_ledger_pool[
                ledger_config.get("id") or ledger_config.get("pool_name")
            ] = ledger_config
    if "ledger.write_ledger" in settings:
        if settings.get("ledger.write_ledger") in prod_write_ledger_pool:
            write_ledger_config = prod_write_ledger_pool.get(
                settings.get("ledger.write_ledger")
            )
        elif settings.get("ledger.write_ledger") in non_prod_write_ledger_pool:
            write_ledger_config = non_prod_write_ledger_pool.get(
                settings.get("ledger.write_ledger")
            )
        else:
            raise ProfileError(
                "The ledger.write_ledger in profile settings does not correspond to a"
                " write configurable ledger provided with --genesis-transactions-list"
            )
    else:
        if len(prod_write_ledger_pool) >= 1:
            write_ledger_config = (list(prod_write_ledger_pool.values()))[0]
        elif len(non_prod_write_ledger_pool) >= 1:
            write_ledger_config = (list(non_prod_write_ledger_pool.values()))[0]
        else:
            raise ProfileError(
                "No write ledger configuration found in ledger_config_list which "
                "was provided with --genesis-transactions-list"
            )
    return write_ledger_config



"""Represents an invitation message for establishing connection."""

from typing import Optional, Sequence
from urllib.parse import parse_qs, urljoin, urlparse

from marshmallow import EXCLUDE, ValidationError, fields, pre_load, validates_schema

from .....did.did_key import DIDKey
from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.valid import (
    GENERIC_DID_EXAMPLE,
    GENERIC_DID_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
)
from .....wallet.util import b64_to_bytes, bytes_to_b64
from ..message_types import CONNECTION_INVITATION, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".connection_invitation_handler.ConnectionInvitationHandler"
)


class ConnectionInvitation(AgentMessage):
    """Class representing a connection invitation."""

    class Meta:
        """Metadata for a connection invitation."""

        handler_class = HANDLER_CLASS
        message_type = CONNECTION_INVITATION
        schema_class = "ConnectionInvitationSchema"

    def __init__(
        self,
        *,
        label: Optional[str] = None,
        did: Optional[str] = None,
        recipient_keys: Sequence[str] = None,
        endpoint: Optional[str] = None,
        routing_keys: Sequence[str] = None,
        image_url: Optional[str] = None,
        **kwargs,
    ):
        """Initialize connection invitation object.

        Args:
            label: Optional label for connection invitation
            did: DID for this connection invitation
            recipient_keys: List of recipient keys
            endpoint: Endpoint which this agent can be reached at
            routing_keys: List of routing keys
            image_url: Optional image URL for connection invitation
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.label = label
        self.did = did
        self.recipient_keys = list(recipient_keys) if recipient_keys else None
        self.endpoint = endpoint
        self.routing_keys = list(routing_keys) if routing_keys else None
        self.routing_keys = (
            [
                (
                    DIDKey.from_did(key).public_key_b58
                    if key.startswith("did:key:")
                    else key
                )
                for key in self.routing_keys
            ]
            if self.routing_keys
            else None
        )
        self.image_url = image_url

    def to_url(self, base_url: Optional[str] = None) -> str:
        """Convert an invitation to URL format for sharing.

        Returns:
            An invite url

        """
        c_json = self.to_json()
        c_i = bytes_to_b64(c_json.encode("ascii"), urlsafe=True, pad=False)
        result = urljoin(base_url or self.endpoint or "", "?c_i={}".format(c_i))
        return result

    @classmethod
    def from_url(cls, url: str) -> "ConnectionInvitation":
        """Parse a URL-encoded invitation into a `ConnectionInvitation` message.

        Args:
            url: Url to decode

        Returns:
            A `ConnectionInvitation` object.

        """
        parts = urlparse(url)
        query = parse_qs(parts.query)
        if "c_i" in query:
            c_i = b64_to_bytes(query["c_i"][0], urlsafe=True)
            return cls.from_json(c_i)
        return None


class ConnectionInvitationSchema(AgentMessageSchema):
    """Connection invitation schema class."""

    class Meta:
        """Connection invitation schema metadata."""

        model_class = ConnectionInvitation
        unknown = EXCLUDE

    label = fields.Str(
        required=False,
        metadata={
            "description": "Optional label for connection invitation",
            "example": "Bob",
        },
    )
    did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={
            "description": "DID for connection invitation",
            "example": GENERIC_DID_EXAMPLE,
        },
    )
    recipient_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Recipient public key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        data_key="recipientKeys",
        required=False,
        metadata={"description": "List of recipient keys"},
    )
    endpoint = fields.Str(
        data_key="serviceEndpoint",
        required=False,
        metadata={
            "description": "Service endpoint at which to reach this agent",
            "example": "http://192.168.56.101:8020",
        },
    )
    routing_keys = fields.List(
        fields.Str(
            validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
            metadata={
                "description": "Routing key",
                "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
            },
        ),
        data_key="routingKeys",
        required=False,
        metadata={"description": "List of routing keys"},
    )
    image_url = fields.URL(
        data_key="imageUrl",
        required=False,
        allow_none=True,
        metadata={
            "description": "Optional image URL for connection invitation",
            "example": "http://192.168.56.101/img/logo.jpg",
        },
    )

    @pre_load
    def transform_routing_keys(self, data, **kwargs):
        """Transform routingKeys from did:key refs, if necessary."""
        routing_keys = data.get("routingKeys")
        if routing_keys:
            data["routingKeys"] = [
                (
                    DIDKey.from_did(key).public_key_b58
                    if key.startswith("did:key:")
                    else key
                )
                for key in routing_keys
            ]
        return data

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: If any of the fields do not validate

        """
        if data.get("did"):
            if data.get("recipient_keys"):
                raise ValidationError("Fields are incompatible", ("did", "recipientKeys"))
            if data.get("endpoint"):
                raise ValidationError(
                    "Fields are incompatible", ("did", "serviceEndpoint")
                )
        elif not data.get("recipient_keys") or not data.get("endpoint"):
            raise ValidationError(
                "Missing required field(s)", ("did", "recipientKeys", "serviceEndpoint")
            )



"""Wallet configuration."""

import logging
from typing import Tuple

from ..core.error import ProfileNotFoundError
from ..core.profile import Profile, ProfileManager, ProfileSession
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..version import RECORD_TYPE_ACAPY_VERSION, __version__
from ..wallet.base import BaseWallet
from ..wallet.crypto import seed_to_did
from ..wallet.did_info import DIDInfo
from ..wallet.did_method import SOV
from ..wallet.key_type import ED25519
from .base import ConfigError
from .injection_context import InjectionContext

LOGGER = logging.getLogger(__name__)

CFG_MAP = {
    "key",
    "key_derivation_method",
    "rekey",
    "name",
    "storage_config",
    "storage_creds",
    "storage_type",
    "test",
}


async def wallet_config(
    context: InjectionContext, provision: bool = False
) -> Tuple[Profile, DIDInfo]:
    """Initialize the root profile."""

    mgr = context.inject(ProfileManager)

    settings = context.settings
    profile_cfg = {}
    for k in CFG_MAP:
        pk = f"wallet.{k}"
        if pk in settings:
            profile_cfg[k] = settings[pk]

    # may be set by `aca-py provision --recreate`
    if settings.get("wallet.recreate"):
        profile_cfg["auto_recreate"] = True

    if provision:
        profile = await mgr.provision(context, profile_cfg)
    else:
        try:
            profile = await mgr.open(context, profile_cfg)
        except ProfileNotFoundError:
            if settings.get("auto_provision", False):
                profile = await mgr.provision(context, profile_cfg)
                provision = True
            else:
                raise

    if provision:
        if profile.created:
            print("Created new profile")
        else:
            print("Opened existing profile")
        print("Profile backend:", profile.backend)
        print("Profile name:", profile.name)

    wallet_seed = context.settings.get("wallet.seed")
    wallet_local_did = context.settings.get("wallet.local_did")
    txn = await profile.transaction()
    wallet = txn.inject(BaseWallet)

    public_did_info = await wallet.get_public_did()
    public_did = None

    if public_did_info:
        public_did = public_did_info.did
        if wallet_seed and seed_to_did(wallet_seed) != public_did:
            if context.settings.get("wallet.replace_public_did"):
                replace_did_info = await wallet.create_local_did(
                    method=SOV, key_type=ED25519, seed=wallet_seed
                )
                public_did = replace_did_info.did
                await wallet.set_public_did(public_did)
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {replace_did_info.verkey}")
            else:
                # If we already have a registered public did and it doesn't match
                # the one derived from `wallet_seed` then we error out.
                raise ConfigError(
                    "New seed provided which doesn't match the registered"
                    + f" public did {public_did}"
                )
        # wait until ledger config to set public DID endpoint - wallet goes first
    elif wallet_seed:
        if wallet_local_did:
            endpoint = context.settings.get("default_endpoint")
            metadata = {"endpoint": endpoint} if endpoint else None

            local_did_info = await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=wallet_seed,
                metadata=metadata,
            )
            local_did = local_did_info.did
            if provision:
                print(f"Created new local DID: {local_did}")
                print(f"Verkey: {local_did_info.verkey}")
        else:
            public_did_info = await wallet.create_public_did(
                method=SOV, key_type=ED25519, seed=wallet_seed
            )
            public_did = public_did_info.did
            if provision:
                print(f"Created new public DID: {public_did}")
                print(f"Verkey: {public_did_info.verkey}")
            # wait until ledger config to set public DID endpoint - wallet goes first

    if provision and not wallet_local_did and not public_did:
        print("No public DID")

    # Debug settings
    test_seed = context.settings.get("debug.seed")
    if context.settings.get("debug.enabled"):
        if not test_seed:
            test_seed = "testseed000000000000000000000001"
    if test_seed:
        await wallet.create_local_did(
            method=SOV,
            key_type=ED25519,
            seed=test_seed,
            metadata={"endpoint": "1.2.3.4:8021"},
        )

    await txn.commit()

    return (profile, public_did_info)


async def add_or_update_version_to_storage(session: ProfileSession):
    """Add or update ACA-Py version StorageRecord."""
    storage: BaseStorage = session.inject(BaseStorage)
    try:
        record = await storage.find_record(RECORD_TYPE_ACAPY_VERSION, {})
        await storage.update_record(record, f"v{__version__}", {})
    except StorageNotFoundError:
        raise ConfigError(
            (
                "No wallet storage version found, Run aca-py "
                "upgrade command with --from-version argument "
                "to fix this."
            )
        )



from unittest import IsolatedAsyncioTestCase

import pytest

from .....cache.base import BaseCache
from .....cache.in_memory import InMemoryCache
from .....connections.models.conn_record import ConnRecord
from .....connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from .....core.oob_processor import OobMessageProcessor
from .....messaging.responder import BaseResponder, MockResponder
from .....multitenant.base import BaseMultitenantManager
from .....multitenant.manager import MultitenantManager
from .....resolver.default.legacy_peer import LegacyPeerDIDResolver
from .....resolver.did_resolver import DIDResolver
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....transport.inbound.receipt import MessageReceipt
from .....utils.testing import create_test_profile
from .....wallet.askar import AskarWallet
from .....wallet.base import BaseWallet, DIDInfo
from .....wallet.did_method import SOV, DIDMethods
from .....wallet.key_type import ED25519, KeyTypes
from ....coordinate_mediation.v1_0.manager import MediationManager
from ....coordinate_mediation.v1_0.messages.mediate_request import MediationRequest
from ....coordinate_mediation.v1_0.models.mediation_record import MediationRecord
from ....coordinate_mediation.v1_0.route_manager import RouteManager
from ..manager import ConnectionManager, ConnectionManagerError
from ..messages.connection_invitation import ConnectionInvitation
from ..messages.connection_request import ConnectionRequest
from ..messages.connection_response import ConnectionResponse
from ..models.connection_detail import ConnectionDetail


@pytest.mark.filterwarnings("ignore:Aries RFC 0160.*:DeprecationWarning")
class TestConnectionManager(IsolatedAsyncioTestCase):
    def make_did_doc(self, did, verkey):
        doc = DIDDoc(did=did)
        controller = did
        ident = "1"
        pk_value = verkey
        pk = PublicKey(
            did, ident, pk_value, PublicKeyType.ED25519_SIG_2018, controller, False
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            did, "indy", "IndyAgent", recip_keys, router_keys, self.test_endpoint
        )
        doc.set(service)
        return doc

    async def asyncSetUp(self):
        self.test_seed = "testseed000000000000000000000001"
        self.test_did = "55GkHamhTU1ZbTbV2ab9DE"
        self.test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
        self.test_endpoint = "http://localhost"

        self.test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
        self.test_target_verkey = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"

        self.responder = MockResponder()

        self.oob_mock = mock.MagicMock(OobMessageProcessor, autospec=True)
        self.oob_mock.clean_finished_oob_record = mock.CoroutineMock(return_value=None)
        self.route_manager = mock.MagicMock(RouteManager)
        self.route_manager.routing_info = mock.CoroutineMock(
            return_value=([], self.test_endpoint)
        )
        self.route_manager.mediation_record_if_id = mock.CoroutineMock(return_value=None)
        self.resolver = DIDResolver()
        self.resolver.register_resolver(LegacyPeerDIDResolver())

        self.profile = await create_test_profile(
            {
                "default_endpoint": "http://aries.ca/endpoint",
                "default_label": "This guy",
                "additional_endpoints": ["http://aries.ca/another-endpoint"],
                "debug.auto_accept_invites": True,
                "debug.auto_accept_requests": True,
            },
        )

        self.profile.context.injector.bind_instance(BaseResponder, self.responder)
        self.profile.context.injector.bind_instance(BaseCache, InMemoryCache())
        self.profile.context.injector.bind_instance(OobMessageProcessor, self.oob_mock)
        self.profile.context.injector.bind_instance(RouteManager, self.route_manager)
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.profile.context.injector.bind_instance(DIDResolver, self.resolver)
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())
        self.context = self.profile.context

        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.context.injector.bind_instance(BaseMultitenantManager, self.multitenant_mgr)

        self.test_mediator_routing_keys = ["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRR"]
        self.test_mediator_conn_id = "mediator-conn-id"
        self.test_mediator_endpoint = "http://mediator.example.com"

        self.manager = ConnectionManager(self.profile)
        assert self.manager.profile

    async def test_create_invitation_non_multi_use_invitation_fails_on_reuse(self):
        connect_record, _ = await self.manager.create_invitation()

        receipt = MessageReceipt(recipient_verkey=connect_record.invitation_key)

        request_a = ConnectionRequest(
            connection=ConnectionDetail(
                did=self.test_target_did,
                did_doc=self.make_did_doc(self.test_target_did, self.test_target_verkey),
            ),
            label="SameInviteRequestA",
        )

        await self.manager.receive_request(request_a, receipt)

        request_b = ConnectionRequest(
            connection=ConnectionDetail(
                did=self.test_did,
                did_doc=self.make_did_doc(self.test_did, self.test_verkey),
            ),
            label="SameInviteRequestB",
        )

        # requestB fails because the invitation was not set to multi-use
        with self.assertRaises(ConnectionManagerError):
            await self.manager.receive_request(request_b, receipt)

    async def test_create_invitation_public(self):
        self.context.update_settings({"public_invites": True})

        self.route_manager.route_verkey = mock.CoroutineMock()
        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            connect_record, connect_invite = await self.manager.create_invitation(
                public=True, my_endpoint="testendpoint"
            )

            assert connect_record
            assert connect_invite.did.endswith(self.test_did)
            self.route_manager.route_verkey.assert_called_once_with(
                self.profile, self.test_verkey
            )

    async def test_create_invitation_public_no_public_invites(self):
        self.context.update_settings({"public_invites": False})

        with self.assertRaises(ConnectionManagerError):
            await self.manager.create_invitation(public=True, my_endpoint="testendpoint")

    async def test_create_invitation_public_no_public_did(self):
        self.context.update_settings({"public_invites": True})

        with mock.patch.object(
            AskarWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = None
            with self.assertRaises(ConnectionManagerError):
                await self.manager.create_invitation(
                    public=True, my_endpoint="testendpoint"
                )

    async def test_create_invitation_multi_use(self):
        connect_record, _ = await self.manager.create_invitation(
            my_endpoint="testendpoint", multi_use=True
        )

        receipt = MessageReceipt(recipient_verkey=connect_record.invitation_key)

        request_a = ConnectionRequest(
            connection=ConnectionDetail(
                did=self.test_target_did,
                did_doc=self.make_did_doc(self.test_target_did, self.test_target_verkey),
            ),
            label="SameInviteRequestA",
        )

        await self.manager.receive_request(request_a, receipt)

        request_b = ConnectionRequest(
            connection=ConnectionDetail(
                did=self.test_did,
                did_doc=self.make_did_doc(self.test_did, self.test_verkey),
            ),
            label="SameInviteRequestB",
        )

        await self.manager.receive_request(request_b, receipt)

    async def test_create_invitation_recipient_routing_endpoint(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=self.test_seed,
                did=self.test_did,
                metadata=None,
            )
            connect_record, _ = await self.manager.create_invitation(
                my_endpoint=self.test_endpoint,
                recipient_keys=[self.test_verkey],
                routing_keys=[self.test_verkey],
            )

            receipt = MessageReceipt(recipient_verkey=connect_record.invitation_key)

            request_a = ConnectionRequest(
                connection=ConnectionDetail(
                    did=self.test_target_did,
                    did_doc=self.make_did_doc(
                        self.test_target_did, self.test_target_verkey
                    ),
                ),
                label="InviteRequestA",
            )

            await self.manager.receive_request(request_a, receipt)

    async def test_create_invitation_metadata_assigned(self):
        async with self.profile.session() as session:
            record, _ = await self.manager.create_invitation(metadata={"hello": "world"})

            assert await record.metadata_get_all(session) == {"hello": "world"}

    async def test_create_invitation_multi_use_metadata_transfers_to_connection(self):
        async with self.profile.session() as session:
            connect_record, _ = await self.manager.create_invitation(
                my_endpoint="testendpoint", multi_use=True, metadata={"test": "value"}
            )

            receipt = MessageReceipt(recipient_verkey=connect_record.invitation_key)

            request = ConnectionRequest(
                connection=ConnectionDetail(
                    did=self.test_target_did,
                    did_doc=self.make_did_doc(
                        self.test_target_did, self.test_target_verkey
                    ),
                ),
                label="request",
            )

            new_conn_rec = await self.manager.receive_request(request, receipt)
            assert new_conn_rec != connect_record
            assert await new_conn_rec.metadata_get_all(session) == {"test": "value"}

    async def test_create_invitation_mediation_overwrites_routing_and_endpoint(self):
        self.route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        async with self.profile.session() as session:
            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)
            with mock.patch.object(
                MediationManager,
                "get_default_mediator",
            ) as mock_get_default_mediator:
                _, invite = await self.manager.create_invitation(
                    routing_keys=[self.test_verkey],
                    my_endpoint=self.test_endpoint,
                    mediation_id=mediation_record.mediation_id,
                )
                assert invite.routing_keys == self.test_mediator_routing_keys
                assert invite.endpoint == self.test_mediator_endpoint
                mock_get_default_mediator.assert_not_called()

    async def test_create_invitation_mediation_using_default(self):
        self.route_manager.routing_info = mock.CoroutineMock(
            return_value=(self.test_mediator_routing_keys, self.test_mediator_endpoint)
        )
        async with self.profile.session() as session:
            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)
            with mock.patch.object(
                self.route_manager,
                "mediation_record_if_id",
                mock.CoroutineMock(return_value=mediation_record),
            ):
                _, invite = await self.manager.create_invitation(
                    routing_keys=[self.test_verkey],
                    my_endpoint=self.test_endpoint,
                )
                assert invite.routing_keys == self.test_mediator_routing_keys
                assert invite.endpoint == self.test_mediator_endpoint
                self.route_manager.routing_info.assert_awaited_once_with(
                    self.profile, mediation_record
                )

    async def test_receive_invitation(self):
        (_, connect_invite) = await self.manager.create_invitation(
            my_endpoint="testendpoint"
        )

        invitee_record = await self.manager.receive_invitation(connect_invite)
        assert ConnRecord.State.get(invitee_record.state) is ConnRecord.State.REQUEST

    async def test_receive_invitation_no_auto_accept(self):
        (_, connect_invite) = await self.manager.create_invitation(
            my_endpoint="testendpoint"
        )

        invitee_record = await self.manager.receive_invitation(
            connect_invite, auto_accept=False
        )
        assert ConnRecord.State.get(invitee_record.state) is ConnRecord.State.INVITATION

    async def test_receive_invitation_bad_invitation(self):
        x_invites = [
            ConnectionInvitation(),
            ConnectionInvitation(
                recipient_keys=["3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"]
            ),
        ]

        for x_invite in x_invites:
            with self.assertRaises(ConnectionManagerError):
                await self.manager.receive_invitation(x_invite)

    async def test_receive_invitation_with_did(self):
        """Test invitation received with a public DID instead of service info."""
        invite = ConnectionInvitation(did=self.test_did)
        invitee_record = await self.manager.receive_invitation(invite)
        assert ConnRecord.State.get(invitee_record.state) is ConnRecord.State.REQUEST

    async def test_receive_invitation_mediation_passes_id_when_auto_accept(self):
        with mock.patch.object(ConnectionManager, "create_request") as create_request:
            _, connect_invite = await self.manager.create_invitation(
                my_endpoint="testendpoint"
            )

            invitee_record = await self.manager.receive_invitation(
                connect_invite, mediation_id="test-mediation-id", auto_accept=True
            )
            create_request.assert_called_once_with(
                invitee_record, mediation_id="test-mediation-id"
            )

    async def test_create_request(self):
        conn_req = await self.manager.create_request(
            ConnRecord(
                invitation_key=self.test_verkey,
                their_label="Hello",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                alias="Bob",
            )
        )
        assert conn_req

    async def test_create_request_my_endpoint(self):
        conn_req = await self.manager.create_request(
            ConnRecord(
                invitation_key=self.test_verkey,
                their_label="Hello",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                alias="Bob",
            ),
            my_endpoint="http://testendpoint.com/endpoint",
        )
        assert conn_req

    async def test_create_request_my_did(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )
            conn_req = await self.manager.create_request(
                ConnRecord(
                    invitation_key=self.test_verkey,
                    my_did=self.test_did,
                    their_label="Hello",
                    their_role=ConnRecord.Role.RESPONDER.rfc160,
                    alias="Bob",
                )
            )
            assert conn_req

    async def test_create_request_multitenant(self):
        self.context.update_settings(
            {"wallet.id": "test_wallet", "multitenant.enabled": True}
        )
        mediation_record = MediationRecord(
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )

        with (
            mock.patch.object(
                AskarWallet, "create_local_did", autospec=True
            ) as mock_wallet_create_local_did,
            mock.patch.object(
                ConnectionManager, "create_did_document", autospec=True
            ) as create_did_document,
            mock.patch.object(
                self.route_manager,
                "mediation_records_for_connection",
                mock.CoroutineMock(return_value=[mediation_record]),
            ),
        ):
            mock_wallet_create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            await self.manager.create_request(
                ConnRecord(
                    invitation_key=self.test_verkey,
                    their_label="Hello",
                    their_role=ConnRecord.Role.RESPONDER.rfc160,
                    alias="Bob",
                ),
                my_endpoint=self.test_endpoint,
            )
            create_did_document.assert_called_once_with(
                self.manager,
                mock_wallet_create_local_did.return_value,
                [self.test_endpoint],
                mediation_records=[mediation_record],
            )
            self.route_manager.route_connection_as_invitee.assert_called_once()

    async def test_create_request_mediation_id(self):
        mediation_record = MediationRecord(
            mediation_id="test_mediation_id",
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )

        record = ConnRecord(
            invitation_key=self.test_verkey,
            their_label="Hello",
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            alias="Bob",
        )

        # Ensure the path with new did creation is hit
        record.my_did = None

        with (
            mock.patch.object(
                ConnectionManager, "create_did_document", autospec=True
            ) as create_did_document,
            mock.patch.object(AskarWallet, "create_local_did") as create_local_did,
            mock.patch.object(
                self.route_manager,
                "mediation_records_for_connection",
                mock.CoroutineMock(return_value=[mediation_record]),
            ),
        ):
            did_info = DIDInfo(
                did=self.test_did,
                verkey=self.test_verkey,
                metadata={},
                method=SOV,
                key_type=ED25519,
            )
            create_local_did.return_value = did_info
            await self.manager.create_request(
                record,
                mediation_id=mediation_record.mediation_id,
                my_endpoint=self.test_endpoint,
            )
            create_local_did.assert_called_once_with(SOV, ED25519)
            create_did_document.assert_called_once_with(
                self.manager,
                did_info,
                [self.test_endpoint],
                mediation_records=[mediation_record],
            )

    async def test_create_request_default_mediator(self):
        async with self.profile.session() as session:
            mediation_record = MediationRecord(
                role=MediationRecord.ROLE_CLIENT,
                state=MediationRecord.STATE_GRANTED,
                connection_id=self.test_mediator_conn_id,
                routing_keys=self.test_mediator_routing_keys,
                endpoint=self.test_mediator_endpoint,
            )
            await mediation_record.save(session)

            record = ConnRecord(
                invitation_key=self.test_verkey,
                their_label="Hello",
                their_role=ConnRecord.Role.RESPONDER.rfc160,
                alias="Bob",
            )

            # Ensure the path with new did creation is hit
            record.my_did = None

            with (
                mock.patch.object(
                    ConnectionManager, "create_did_document", autospec=True
                ) as create_did_document,
                mock.patch.object(AskarWallet, "create_local_did") as create_local_did,
                mock.patch.object(
                    self.route_manager,
                    "mediation_records_for_connection",
                    mock.CoroutineMock(return_value=[mediation_record]),
                ),
            ):
                did_info = DIDInfo(
                    did=self.test_did,
                    verkey=self.test_verkey,
                    metadata={},
                    method=SOV,
                    key_type=ED25519,
                )
                create_local_did.return_value = did_info
                await self.manager.create_request(
                    record,
                    my_endpoint=self.test_endpoint,
                )
                create_local_did.assert_called_once_with(SOV, ED25519)
                create_did_document.assert_called_once_with(
                    self.manager,
                    did_info,
                    [self.test_endpoint],
                    mediation_records=[mediation_record],
                )

    async def test_receive_request_public_did_oob_invite(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = self.test_did

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings({"public_invites": True})
            with (
                mock.patch.object(ConnRecord, "connection_id", autospec=True),
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
                mock.patch.object(
                    ConnRecord, "retrieve_by_invitation_msg_id", mock.CoroutineMock()
                ) as mock_conn_retrieve_by_invitation_msg_id,
                mock.patch.object(
                    self.manager, "store_did_document", mock.CoroutineMock()
                ),
            ):
                mock_conn_retrieve_by_invitation_msg_id.return_value = ConnRecord()
                conn_rec = await self.manager.receive_request(mock_request, receipt)
                assert conn_rec

                self.oob_mock.clean_finished_oob_record.assert_called_once_with(
                    self.profile, mock_request
                )

    async def test_receive_request_public_did_unsolicited_fails(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = self.test_did

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings({"public_invites": True})
            with (
                self.assertRaises(ConnectionManagerError),
                mock.patch.object(ConnRecord, "connection_id", autospec=True),
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
                mock.patch.object(
                    ConnRecord, "retrieve_by_invitation_msg_id", mock.CoroutineMock()
                ) as mock_conn_retrieve_by_invitation_msg_id,
                mock.patch.object(
                    self.manager, "store_did_document", mock.CoroutineMock()
                ),
            ):
                mock_conn_retrieve_by_invitation_msg_id.return_value = None
                await self.manager.receive_request(mock_request, receipt)

    async def test_receive_request_public_did_conn_invite(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = self.test_did

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            mock_connection_record = mock.MagicMock()
            mock_connection_record.save = mock.CoroutineMock()
            mock_connection_record.attach_request = mock.CoroutineMock()

            self.context.update_settings({"public_invites": True})
            with (
                mock.patch.object(ConnRecord, "connection_id", autospec=True),
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
                mock.patch.object(
                    ConnRecord,
                    "retrieve_by_invitation_msg_id",
                    mock.CoroutineMock(return_value=mock_connection_record),
                ),
                mock.patch.object(
                    self.manager, "store_did_document", mock.CoroutineMock()
                ),
            ):
                conn_rec = await self.manager.receive_request(mock_request, receipt)
                assert conn_rec

    async def test_receive_request_public_did_unsolicited(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = self.test_did

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings({"public_invites": True})
            self.context.update_settings({"requests_through_public_did": True})
            with (
                mock.patch.object(ConnRecord, "connection_id", autospec=True),
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
                mock.patch.object(
                    ConnRecord, "retrieve_by_invitation_msg_id", mock.CoroutineMock()
                ) as mock_conn_retrieve_by_invitation_msg_id,
                mock.patch.object(
                    self.manager, "store_did_document", mock.CoroutineMock()
                ),
            ):
                mock_conn_retrieve_by_invitation_msg_id.return_value = None
                conn_rec = await self.manager.receive_request(mock_request, receipt)
                assert conn_rec

    async def test_receive_request_public_did_no_did_doc(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = None

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings({"public_invites": True})
            with (
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            ):
                with self.assertRaises(ConnectionManagerError):
                    await self.manager.receive_request(mock_request, receipt)

    async def test_receive_request_public_did_wrong_did(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = "dummy"

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings({"public_invites": True})
            with (
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            ):
                with self.assertRaises(ConnectionManagerError):
                    await self.manager.receive_request(mock_request, receipt)

    async def test_receive_request_public_did_no_public_invites(self):
        mock_request = mock.MagicMock()
        mock_request.connection = mock.MagicMock()
        mock_request.connection.did = self.test_did
        mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_request.connection.did_doc.did = self.test_did

        receipt = MessageReceipt(recipient_did=self.test_did, recipient_did_public=True)
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

        self.context.update_settings({"public_invites": False})
        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(ConnRecord, "attach_request", autospec=True),
            mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
            mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            mock.patch.object(self.manager, "store_did_document", mock.CoroutineMock()),
        ):
            with self.assertRaises(ConnectionManagerError):
                await self.manager.receive_request(mock_request, receipt)

    async def test_receive_request_public_did_no_auto_accept(self):
        async with self.profile.session() as session:
            mock_request = mock.MagicMock()
            mock_request.connection = mock.MagicMock()
            mock_request.connection.did = self.test_did
            mock_request.connection.did_doc = mock.MagicMock(spec=DIDDoc)
            mock_request.connection.did_doc.did = self.test_did

            receipt = MessageReceipt(
                recipient_did=self.test_did, recipient_did_public=True
            )
            wallet = session.inject(BaseWallet)
            await wallet.create_local_did(
                method=SOV,
                key_type=ED25519,
                seed=None,
                did=self.test_did,
            )

            self.context.update_settings(
                {"public_invites": True, "debug.auto_accept_requests": False}
            )
            with (
                mock.patch.object(ConnRecord, "save", autospec=True),
                mock.patch.object(ConnRecord, "attach_request", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_by_id", autospec=True),
                mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
                mock.patch.object(
                    ConnRecord, "retrieve_by_invitation_msg_id", mock.CoroutineMock()
                ) as mock_conn_retrieve_by_invitation_msg_id,
                mock.patch.object(
                    self.manager, "store_did_document", mock.CoroutineMock()
                ),
            ):
                mock_conn_retrieve_by_invitation_msg_id.return_value = ConnRecord()
                conn_rec = await self.manager.receive_request(mock_request, receipt)
                assert conn_rec

            messages = self.responder.messages
            assert not messages

    async def test_create_response(self):
        conn_rec = ConnRecord(state=ConnRecord.State.REQUEST.rfc160)

        with (
            mock.patch.object(ConnRecord, "log_state", autospec=True),
            mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(ConnectionResponse, "sign_field", autospec=True),
            mock.patch.object(conn_rec, "metadata_get", mock.CoroutineMock()),
        ):
            await self.manager.create_response(conn_rec, "http://10.20.30.40:5060/")

    async def test_create_response_multitenant(self):
        self.context.update_settings(
            {"wallet.id": "test_wallet", "multitenant.enabled": True}
        )

        mediation_record = MediationRecord(
            mediation_id="test_mediation_id",
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )

        with (
            mock.patch.object(ConnRecord, "log_state", autospec=True),
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "metadata_get", mock.CoroutineMock(return_value=False)
            ),
            mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            mock.patch.object(ConnectionResponse, "sign_field", autospec=True),
            mock.patch.object(
                AskarWallet, "create_local_did", autospec=True
            ) as mock_wallet_create_local_did,
            mock.patch.object(
                ConnectionManager, "create_did_document", autospec=True
            ) as create_did_document,
            mock.patch.object(
                self.route_manager,
                "mediation_records_for_connection",
                mock.CoroutineMock(return_value=[mediation_record]),
            ),
        ):
            mock_wallet_create_local_did.return_value = DIDInfo(
                self.test_did,
                self.test_verkey,
                None,
                method=SOV,
                key_type=ED25519,
            )
            await self.manager.create_response(
                ConnRecord(
                    state=ConnRecord.State.REQUEST,
                ),
                my_endpoint=self.test_endpoint,
            )
            create_did_document.assert_called_once_with(
                self.manager,
                mock_wallet_create_local_did.return_value,
                [self.test_endpoint],
                mediation_records=[mediation_record],
            )
            self.route_manager.route_connection_as_inviter.assert_called_once()

    async def test_create_response_bad_state(self):
        with self.assertRaises(ConnectionManagerError):
            await self.manager.create_response(
                ConnRecord(
                    invitation_key=self.test_verkey,
                    their_label="Hello",
                    their_role=ConnRecord.Role.RESPONDER.rfc160,
                    alias="Bob",
                    state=ConnRecord.State.ABANDONED.rfc160,
                )
            )

    async def test_create_response_mediation(self):
        mediation_record = MediationRecord(
            mediation_id="test_mediation_id",
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )

        record = ConnRecord(
            connection_id="test-conn-id",
            invitation_key=self.test_verkey,
            their_label="Hello",
            their_role=ConnRecord.Role.RESPONDER.rfc160,
            alias="Bob",
            state=ConnRecord.State.REQUEST.rfc160,
        )

        # Ensure the path with new did creation is hit
        record.my_did = None

        with (
            mock.patch.object(ConnRecord, "log_state", autospec=True),
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                record, "metadata_get", mock.CoroutineMock(return_value=False)
            ),
            mock.patch.object(
                ConnectionManager, "create_did_document", autospec=True
            ) as create_did_document,
            mock.patch.object(AskarWallet, "create_local_did") as create_local_did,
            mock.patch.object(
                self.route_manager,
                "mediation_records_for_connection",
                mock.CoroutineMock(return_value=[mediation_record]),
            ),
            mock.patch.object(record, "retrieve_request", autospec=True),
            mock.patch.object(ConnectionResponse, "sign_field", autospec=True),
        ):
            did_info = DIDInfo(
                did=self.test_did,
                verkey=self.test_verkey,
                metadata={},
                method=SOV,
                key_type=ED25519,
            )
            create_local_did.return_value = did_info
            await self.manager.create_response(
                record,
                mediation_id=mediation_record.mediation_id,
                my_endpoint=self.test_endpoint,
            )
            create_local_did.assert_called_once_with(SOV, ED25519)
            create_did_document.assert_called_once_with(
                self.manager,
                did_info,
                [self.test_endpoint],
                mediation_records=[mediation_record],
            )
            self.route_manager.route_connection_as_inviter.assert_called_once()

    async def test_create_response_auto_send_mediation_request(self):
        conn_rec = ConnRecord(
            state=ConnRecord.State.REQUEST.rfc160,
        )
        conn_rec.my_did = None

        with (
            mock.patch.object(ConnRecord, "log_state", autospec=True),
            mock.patch.object(ConnRecord, "retrieve_request", autospec=True),
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(ConnectionResponse, "sign_field", autospec=True),
            mock.patch.object(
                conn_rec, "metadata_get", mock.CoroutineMock(return_value=True)
            ),
        ):
            await self.manager.create_response(conn_rec)

        assert len(self.responder.messages) == 1
        message, target = self.responder.messages[0]
        assert isinstance(message, MediationRequest)
        assert target["connection_id"] == conn_rec.connection_id

    async def test_accept_response_find_by_thread_id(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did
        mock_response.verify_signed_field = mock.CoroutineMock(return_value="sig_verkey")
        receipt = MessageReceipt(recipient_did=self.test_did, recipient_did_public=True)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
            mock.patch.object(
                MediationManager, "get_default_mediator", mock.CoroutineMock()
            ),
            mock.patch.object(self.manager, "store_did_document", mock.CoroutineMock()),
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
                metadata_get=mock.CoroutineMock(),
                connection_id="test-conn-id",
                invitation_key="test-invitation-key",
            )
            conn_rec = await self.manager.accept_response(mock_response, receipt)
            assert conn_rec.their_did == self.test_target_did
            assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.RESPONSE

    async def test_accept_response_not_found_by_thread_id_receipt_has_sender_did(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did
        mock_response.verify_signed_field = mock.CoroutineMock(return_value="sig_verkey")

        receipt = MessageReceipt(sender_did=self.test_target_did)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
            mock.patch.object(
                ConnRecord, "retrieve_by_did", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_did,
            mock.patch.object(
                MediationManager, "get_default_mediator", mock.CoroutineMock()
            ),
            mock.patch.object(self.manager, "store_did_document", mock.CoroutineMock()),
        ):
            mock_conn_retrieve_by_req_id.side_effect = StorageNotFoundError()
            mock_conn_retrieve_by_did.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
                metadata_get=mock.CoroutineMock(return_value=False),
                connection_id="test-conn-id",
                invitation_key="test-invitation-id",
            )

            conn_rec = await self.manager.accept_response(mock_response, receipt)
            assert conn_rec.their_did == self.test_target_did
            assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.RESPONSE

            assert not self.responder.messages

    async def test_accept_response_not_found_by_thread_id_nor_receipt_sender_did(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did

        receipt = MessageReceipt(sender_did=self.test_target_did)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
            mock.patch.object(
                ConnRecord, "retrieve_by_did", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_did,
        ):
            mock_conn_retrieve_by_req_id.side_effect = StorageNotFoundError()
            mock_conn_retrieve_by_did.side_effect = StorageNotFoundError()

            with self.assertRaises(ConnectionManagerError):
                await self.manager.accept_response(mock_response, receipt)

    async def test_accept_response_find_by_thread_id_bad_state(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did

        receipt = MessageReceipt(sender_did=self.test_target_did)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                state=ConnRecord.State.ABANDONED.rfc23
            )

            with self.assertRaises(ConnectionManagerError):
                await self.manager.accept_response(mock_response, receipt)

    async def test_accept_response_find_by_thread_id_no_connection_did_doc(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = None

        receipt = MessageReceipt(sender_did=self.test_target_did)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
            )

            with self.assertRaises(ConnectionManagerError):
                await self.manager.accept_response(mock_response, receipt)

    async def test_accept_response_find_by_thread_id_did_mismatch(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_did

        receipt = MessageReceipt(sender_did=self.test_target_did)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
            )

            with self.assertRaises(ConnectionManagerError):
                await self.manager.accept_response(mock_response, receipt)

    async def test_accept_response_verify_invitation_key_sign_failure(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did
        mock_response.verify_signed_field = mock.CoroutineMock(side_effect=ValueError)
        receipt = MessageReceipt(recipient_did=self.test_did, recipient_did_public=True)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
            mock.patch.object(
                MediationManager, "get_default_mediator", mock.CoroutineMock()
            ),
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
                metadata_get=mock.CoroutineMock(),
                connection_id="test-conn-id",
                invitation_key="test-invitation-key",
            )
            with self.assertRaises(ConnectionManagerError):
                await self.manager.accept_response(mock_response, receipt)

    async def test_accept_response_auto_send_mediation_request(self):
        mock_response = mock.MagicMock()
        mock_response._thread = mock.MagicMock()
        mock_response.connection = mock.MagicMock()
        mock_response.connection.did = self.test_target_did
        mock_response.connection.did_doc = mock.MagicMock(spec=DIDDoc)
        mock_response.connection.did_doc.did = self.test_target_did
        mock_response.verify_signed_field = mock.CoroutineMock(return_value="sig_verkey")
        receipt = MessageReceipt(recipient_did=self.test_did, recipient_did_public=True)

        with (
            mock.patch.object(ConnRecord, "save", autospec=True),
            mock.patch.object(
                ConnRecord, "retrieve_by_request_id", mock.CoroutineMock()
            ) as mock_conn_retrieve_by_req_id,
            mock.patch.object(
                MediationManager, "get_default_mediator", mock.CoroutineMock()
            ),
            mock.patch.object(self.manager, "store_did_document", mock.CoroutineMock()),
        ):
            mock_conn_retrieve_by_req_id.return_value = mock.MagicMock(
                did=self.test_target_did,
                did_doc=mock.MagicMock(did=self.test_target_did),
                state=ConnRecord.State.RESPONSE.rfc23,
                save=mock.CoroutineMock(),
                metadata_get=mock.CoroutineMock(return_value=True),
                connection_id="test-conn-id",
                invitation_key="test-invitation-key",
            )
            conn_rec = await self.manager.accept_response(mock_response, receipt)
            assert conn_rec.their_did == self.test_target_did
            assert ConnRecord.State.get(conn_rec.state) is ConnRecord.State.RESPONSE

            assert len(self.responder.messages) == 1
            message, target = self.responder.messages[0]
            assert isinstance(message, MediationRequest)
            assert target["connection_id"] == conn_rec.connection_id



"""Connection request handler."""

from .....connections.models.conn_record import ConnRecord
from .....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from ....coordinate_mediation.v1_0.manager import MediationManager
from ..manager import ConnectionManager, ConnectionManagerError
from ..messages.connection_request import ConnectionRequest


class ConnectionRequestHandler(BaseHandler):
    """Handler class for connection requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle connection request.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug(f"ConnectionRequestHandler called with context {context}")
        assert isinstance(context.message, ConnectionRequest)

        profile = context.profile
        mgr = ConnectionManager(profile)

        mediation_id = None
        if context.connection_record:
            async with profile.session() as session:
                mediation_metadata = await context.connection_record.metadata_get(
                    session, MediationManager.METADATA_KEY, {}
                )
            mediation_id = mediation_metadata.get(MediationManager.METADATA_ID)

        try:
            connection = await mgr.receive_request(
                context.message,
                context.message_receipt,
            )

            if connection.accept == ConnRecord.ACCEPT_AUTO:
                response = await mgr.create_response(
                    connection, mediation_id=mediation_id
                )
                await responder.send_reply(
                    response, connection_id=connection.connection_id
                )
            else:
                self._logger.debug("Connection request will await acceptance")
        except ConnectionManagerError as e:
            report, targets = mgr.manager_error_to_problem_report(
                e, context.message, context.message_receipt
            )
            if report and targets:
                await responder.send_reply(
                    message=report,
                    target_list=targets,
                )



"""Represents a DID exchange response message under RFC 23."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from .....messaging.valid import GENERIC_DID_EXAMPLE, GENERIC_DID_VALIDATE
from ..message_types import DIDX_RESPONSE, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.response_handler.DIDXResponseHandler"


class DIDXResponse(AgentMessage):
    """Class representing a DID exchange response under RFC 23."""

    class Meta:
        """Metadata for DID exchange response under RFC 23."""

        handler_class = HANDLER_CLASS
        message_type = DIDX_RESPONSE
        schema_class = "DIDXResponseSchema"

    def __init__(
        self,
        *,
        did: Optional[str] = None,
        did_doc_attach: Optional[AttachDecorator] = None,
        did_rotate_attach: Optional[AttachDecorator] = None,
        **kwargs,
    ):
        """Initialize DID exchange response object under RFC 23.

        Args:
            image_url: Optional image URL for this response
            did: DID for this response
            did_doc_attach: signed DID doc attachment
            did_rotate_attach: signed DID rotation attachment
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.did = did
        self.did_doc_attach = did_doc_attach
        self.did_rotate_attach = did_rotate_attach


class DIDXResponseSchema(AgentMessageSchema):
    """Schema class for DID exchange response under RFC 23."""

    class Meta:
        """DID exchange response schema class metadata."""

        model_class = DIDXResponse
        unknown = EXCLUDE

    did = fields.Str(
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of exchange", "example": GENERIC_DID_EXAMPLE},
    )
    did_doc_attach = fields.Nested(
        AttachDecoratorSchema,
        required=False,
        data_key="did_doc~attach",
        metadata={"description": "As signed attachment, DID Doc associated with DID"},
    )
    did_rotate_attach = fields.Nested(
        AttachDecoratorSchema,
        required=False,
        data_key="did_rotate~attach",
        metadata={"description": "As signed attachment, DID signed by invitation key"},
    )



"""Refused transaction response handler."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import TransactionManager, TransactionManagerError
from ..messages.refused_transaction_response import RefusedTransactionResponse


class RefusedTransactionResponseHandler(BaseHandler):
    """Handler class for Refused transaction response."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle refused transaction response.

        Args:
            context: Request context
            responder: Responder callback
        """

        self._logger.debug(
            f"RefusedTransactionResponseHandler called with context {context}"
        )
        assert isinstance(context.message, RefusedTransactionResponse)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        mgr = TransactionManager(context.profile)
        try:
            await mgr.receive_refuse_response(context.message)
        except TransactionManagerError:
            self._logger.exception("Error receiving refused transaction response")



TEST_VC_DOCUMENT_NESTED_REVEAL = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "issuer": {},
    "issuanceDate": {},
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
    "@explicit": True,
    "credentialSubject": {"@explicit": True, "degree": {"@explicit": True, "name": {}}},
}
